#define CARVE_VERSION "2.0.0a"

/* #undef HAVE_TR1_UNORDERED_COLLECTIONS */
/* #undef HAVE_STD_UNORDERED_COLLECTIONS */
/* #undef HAVE_LIBSTDCPP_UNORDERED_COLLECTIONS */
#define HAVE_BOOST_UNORDERED_COLLECTIONS

/* #undef CARVE_SYSTEM_BOOST */

/* #undef CARVE_DEBUG */
/* #undef CARVE_DEBUG_WRITE_PLY_DATA */

#define CARVE_USE_EXACT_PREDICATES
